// Interval arithmetic proposal for C++0x.
//
// Copyright 2005  INRIA, France.
// All rights reserved.
//
// Author : Sylvain Pion.

// Notes :
// -------
// * Possible optimizations :
//   - when double rounding is no issue (e.g. not on x86 without special care),
//     rounding towards -infinity can be most of the time emulated using
//     rounding towards +infinity.  E.g. using rules like : a+b = -(-a-b).
//   - the compiler could be taught about the rounding mode changing functions,
//     and even eliminate rounding mode changes between functions provided
//     they are inlined or inter procedural optimizations are performed.

// TODO :
// * Comparison behavior ?
//   How to not force the exception mecanism, without a policy class ?
//   - Maybe return tribool ?   Simple but requires tribool.
//     Also, sign() cannot return tribool (but we could remove sign()).
//   - Policy ?  Probably too complex for std.
//   - No operators at all ?  Put variants in some namespaces ?
//   - The Sun way is to not provide default operators, but functions
//     for the 3 ways (set, possibly, certainly) => in sub-namespaces ?
// * IO operators.

#ifndef STD_INTERVAL_H_INCLUDED
#define STD_INTERVAL_H_INCLUDED

#include <stdexcept>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <cassert>
#include <cmath>

#include <boost/operators.hpp>
#include <boost/type_traits.hpp>

// #include "interval_bool.h" // included later due to cyclic dependency

namespace std2 {

template < typename T >
class interval
  : boost::field_operators1< interval<T>
  , boost::field_operators2< interval<T>, T > >
{
  T _i, _s;

public:

  typedef T  value_type;

  interval();
  interval(T const& t);
  interval(T const& i, T const& s);
  template < typename U > explicit interval(interval<U> const& i);
  template < typename U > explicit interval(U const& a);
  template < typename U > interval(U const& a, U const& b);

private:
  void assign(T const& a, T const& b);
  void assign_inf(T const& a);
  void assign_sup(T const& a);

public:
  bool empty() const;
  T const& inf() const;
  T const& sup() const;

  interval const& operator+ () const;
  interval        operator- () const;

  interval& operator+= (interval const& i);
  interval& operator-= (interval const& i);
  interval& operator*= (interval const& i);
  interval& operator/= (interval const& i);

  interval& operator+= (T const& t);
  interval& operator-= (T const& t);
  interval& operator*= (T const& t);
  interval& operator/= (T const& t);

  static interval whole();
};

} // namespace std2

// Some internal helpers
// ---------------------


// rounding<T> :
// Helper class for rounding mode selector/protector.
// The default template version does nothing special (aimed at exact types).
// It is specialized for float, double, long double to call <fenv> functions.
// It would be nice to allow the user to specialize it for its own types ?

#include <fenv.h>

namespace std2 {
namespace detail {

template < typename T, bool = boost::is_float<T>::value >
struct rounding
{
  rounding(int) {}
  void set(int) {}
};

template < typename T >
class rounding <T, true>
{
  int backup;
public:
  rounding(int flag) : backup(fegetround())  { fesetround(flag); }
  void set(int flag)                         { fesetround(flag); }
  ~rounding()                                { fesetround(backup); }
};


// Tool to convert between types, respecting containment.
// All specializations are just optimizations...
template < typename T >
struct convert
{
  template < typename U >
  T operator()(U const& t, int = 0) const { return t; }
};

template <>
struct convert <double>
{
  double operator()(long double const& t, int rnd = FE_UPWARD) const
  {
    detail::rounding<double> r(rnd);
    return t;
  }
  double operator()(float       const& t, int = 0) const { return t; }
  double operator()(int         const& t, int rnd = FE_UPWARD) const
  {
    detail::rounding<double> r(rnd);
    return t;
  }
};

template <>
struct convert <float>
{
  float operator()(long double const& t, int rnd = FE_UPWARD) const
  {
    detail::rounding<float> r(rnd);
    return t;
  }
  float operator()(double      const& t, int rnd = FE_UPWARD) const
  {
    detail::rounding<float> r(rnd);
    return t;
  }
  // TODO : consistency wrt integral types...
  float operator()(int         const& t, int rnd = FE_UPWARD) const
  {
    detail::rounding<float> r(rnd);
    return t;
  }
};

} // namespace detail


// Access functions
// ----------------

template < typename T > inline
T const& interval<T>::inf() const
{
  return _i;
}

template < typename T > inline
T const& interval<T>::sup() const
{
  return _s;
}

template < typename T > inline
T const& inf(interval<T> const& i)
{
  return i.inf();
}

template < typename T > inline
T const& sup(interval<T> const& i)
{
  return i.sup();
}


// Special constant intervals
// --------------------------

/* Deprecated: it is now decided that interval() returns an empty interval
 * See interval constructors below
   template < typename T > inline
   interval<T> interval<T>::empty()
   {
     return interval(1, 0);
   }
 */

template < typename T > inline
interval<T> interval<T>::whole()
{
  // we assume that std::numeric_limits<T>::has_infinity
  // implementations not providing infinity need to use DBL_MAX and co.
  T t = std::numeric_limits<T>::infinity();
  return interval(-t, t);
}

// Set operations
// --------------

// NEW: formerly is_empty(), now a member function called empty()
// avoids clash with tr1::is_empty<> which is a struct
template < typename T > inline
bool interval<T>::empty() const
{
  return !(inf() <= sup());
}

template < typename T > inline
bool contains(interval<T> const& i, T const& t)
{
  return inf(i) <= t && t <= sup(i);
}

template < typename T > inline
bool contains(interval<T> const& i1, interval<T> const& i2)
{
  if (i2.empty()) return true;
  if (i1.empty()) return false;
  return inf(i1) <= inf(i2) && sup(i2) <= sup(i1);
}

template < typename T > inline
bool comparable(interval<T> const& i1, interval<T> const& i2)
{
  return !i1.empty() && !i2.empty() &&
         (sup(i1) < inf(i2) || sup(i2) < inf(i1));
}

template < typename T > inline
bool overlap(interval<T> const& i1, interval<T> const& i2)
{
  return !i1.empty() && !i2.empty() &&
         inf(i2) <= sup(i1) && inf(i1) <= sup(i2);
}

template < typename T > inline
bool equal(interval<T> const& i1, interval<T> const& i2)
{
  return ( i1.empty() && i2.empty() ) ||
         ( inf(i1) == inf(i2) && sup(i1) == sup(i2) );
}

template < typename T > inline
bool equal(interval<T> const& i, T const& t)
{
  return inf(i) == t && sup(i) == t;
}

template < typename T > inline
bool equal(T const& t, interval<T> const& i)
{
  return inf(i) == t && sup(i) == t;
}

template < typename T > inline
bool is_singleton(interval<T> const& i)
{
  return inf(i) == sup(i);
}

template < typename T > inline
std::pair<interval<T>, interval<T> > split(interval<T> const& i, T const& t)
{
  if (i.empty())
    return std::make_pair(interval<T>(), interval<T>());
  if (t < i.inf())
    return std::make_pair(interval<T>(), i);
  if (t > i.sup())
    return std::make_pair(i, interval<T>());
  return std::make_pair(interval<T>(inf(i), t),
                        interval<T>(t, sup(i)));
}

template < typename T > inline
std::pair<interval<T>, interval<T> > bisect(interval<T> const& i)
{
  return split(i, midpoint(i));
}

template < typename T > inline
interval<T> intersect(interval<T> const& a, interval<T> const& b)
{
  if (a.empty() || b.empty() ||
      !(inf(b) <= sup(a)) || !(inf(a) <= sup(b)))
    return interval<T>();
  return interval<T>(std::max(inf(a), inf(b)), std::min(sup(a), sup(b)));
}

template < typename T > inline
interval<T> hull(interval<T> const& a, interval<T> const& b)
{
  if (a.empty()) return b;
  if (b.empty()) return a;
  return interval<T>(std::min(inf(a), inf(b)), std::max(sup(a), sup(b)));
}

} // namespace std2


// Comparison operators
// --------------------

// the comparisons implemented here have the following semantic:
// they return an interval<bool> which is either [false], [true], or [true,false]
// the latter value means indetermination (because the intervals overlap)
// interval<bool> is not specified by the C++ standard proposal,
// but also not explicitly disallowed, so we use it here in place
// of a tribool.  We could also have used the boost::tribool library.

#include "interval_bool.h" // included here due to cyclic dependency

namespace std2 {

template < typename T > inline
interval<bool> operator< (interval<T> const& a, interval<T> const& b)
{
  if (a.empty() || b.empty())
    return interval<bool>();

  if (sup(a) <  inf(b)) return true;
  if (inf(a) >= sup(b)) return false;

  return interval<bool>::indeterminate();
}

template < typename T > inline
interval<bool> operator==(interval<T> const& a, interval<T> const& b)
{
  if (a.empty() || b.empty())
    return interval<bool>();

  if (sup(a) == inf(b) && sup(b) == inf(a)) return true;
  if (sup(a) <  inf(b) || sup(b) <  inf(a)) return false;

  return interval<bool>::indeterminate();
}

template < typename T > inline
interval<bool> operator< (interval<T> const& a, T const& b)
{
  if (a.empty())
    return interval<bool>();

  if (sup(a) <  b) return true;
  if (inf(a) >= b) return false;

  return interval<bool>::indeterminate();
}

template < typename T > inline
interval<bool> operator< (T const& a, interval<T> const& b)
{
  if (b.empty())
    return interval<bool>();

  if (a <  inf(b)) return true;
  if (a >= sup(b)) return false;

  return interval<bool>::indeterminate();
}

template < typename T > inline
interval<bool> operator==(interval<T> const& a, T const& b)
{
  if (a.empty())
    return interval<bool>();

  if (sup(a) == b && b == inf(a)) return true;
  if (sup(a) <  b || b <  inf(a)) return false;

  return interval<bool>::indeterminate();
}

// We cannot use boost operators for comparisons since they assume
// bool as return type (I think).

template < typename T > inline
interval<bool> operator> (interval<T> const& a, interval<T> const& b)
{ return b < a; }

template < typename T > inline
interval<bool> operator> (T const& a, interval<T> const& b)
{ return b < a; }

template < typename T > inline
interval<bool> operator> (interval<T> const& a, T const& b)
{ return b < a; }


template < typename T > inline
interval<bool> operator<=(interval<T> const& a, interval<T> const& b)
{ return ! (b < a); }

template < typename T > inline
interval<bool> operator<=(T const& a, interval<T> const& b)
{ return ! (b < a); }

template < typename T > inline
interval<bool> operator<=(interval<T> const& a, T const& b)
{ return ! (b < a); }


template < typename T > inline
interval<bool> operator>=(interval<T> const& a, interval<T> const& b)
{ return ! (a < b); }

template < typename T > inline
interval<bool> operator>=(T const& a, interval<T> const& b)
{ return ! (a < b); }

template < typename T > inline
interval<bool> operator>=(interval<T> const& a, T const& b)
{ return ! (a < b); }


template < typename T > inline
interval<bool> operator!=(interval<T> const& a, interval<T> const& b)
{ return ! (a == b); }

template < typename T > inline
interval<bool> operator!=(T const& a, interval<T> const& b)
{ return ! (a == b); }

template < typename T > inline
interval<bool> operator!=(interval<T> const& a, T const& b)
{ return ! (a == b); }


template < typename T > inline
interval<bool> operator==(T const& a, interval<T> const& b)
{ return b == a; }



// Constructors and assignments
// ----------------------------

template < typename T > inline
interval<T>::interval()
  : _i(T(1)), _s(T(0))
{
}

template < typename T > inline
interval<T>::interval(T const& t)
  : _i(t), _s(t)
{}

template < typename T > inline
interval<T>::interval(T const& a, T const& b)
  : _i(a), _s(b)
{}

template < typename T >
template < typename U > inline
interval<T>::interval(interval<U> const& i)
  : _i(detail::convert<T>()(i.inf(), FE_DOWNWARD)),
    _s(detail::convert<T>()(i.sup()))
{}

template < typename T >
template < typename U > inline
interval<T>::interval(U const& t)
  : _i(detail::convert<T>()(t, FE_DOWNWARD)), _s(detail::convert<T>()(t))
{}

template < typename T >
template < typename U > inline
interval<T>::interval(U const& a, U const& b)
  : _i(detail::convert<T>()(a, FE_DOWNWARD)), _s(detail::convert<T>()(b))
{}

template < typename T > inline
void interval<T>::assign(T const& a, T const& b)
{
  _i = a; _s = b;
}

template < typename T > inline
void interval<T>::assign_inf(T const& a)
{
  _i = a;
}

template < typename T > inline
void interval<T>::assign_sup(T const& a)
{
  _s = a;
}


// Arithmetic functions
// --------------------

// Note: In negation and addition, the tests for empty intervals are not needed
//       if uncomparable numbers are used as bounds of empty intervals.

template < typename T > inline
interval<T> const&
interval<T>::operator+() const
{
  return *this;
}

template < typename T > inline
interval<T>
interval<T>::operator-() const
{
  if (empty())
    return *this;
  return interval(-_s, -_i);
}

template < typename T > inline
interval<T>&
interval<T>::operator+=(interval<T> const& i)
{
  if (!empty())
    if (i.empty())
      *this = empty();
    else {
      detail::rounding<T> r(FE_UPWARD);
      _s += i._s;
      r.set(FE_DOWNWARD);
      _i += i._i;
    }
  return *this;
}

template < typename T > inline
interval<T>&
interval<T>::operator-=(interval<T> const& i)
{
  *this += -i;
  return *this;
}

template < typename T >
interval<T>&
interval<T>::operator*=(interval<T> const& i)
{
  if (empty() || i.empty()) {
    *this = empty();
    return *this;
  }
  detail::rounding<T> r(FE_UPWARD);
  if (inf() >= 0)                                   // this>=0
  {
    // i>=0     [inf()*inf(i); sup()*sup(i)]
    // i<=0     [sup()*inf(i); inf()*sup(i)]
    // i~=0     [sup()*inf(i); sup()*sup(i)]
    T aa = inf(), bb = sup();
    if (i.inf() < 0)
    {
        aa = bb;
        if (i.sup() < 0)
            bb = inf();
    }
    T up = bb * i.sup();
    r.set(FE_DOWNWARD);
    assign(aa * i.inf(), up);
    return *this;
  }
  else if (sup() <= 0)                                // this<=0
  {
    // i>=0     [inf()*sup(i); sup()*inf(i)]
    // i<=0     [sup()*sup(i); inf()*inf(i)]
    // i~=0     [inf()*sup(i); inf()*inf(i)]
    T aa = sup(), bb = inf();
    if (i.inf() < 0)
    {
        aa=bb;
        if (i.sup() < 0)
            bb=sup();
    }
    T up = aa * i.inf();
    r.set(FE_DOWNWARD);
    assign(bb * i.sup(), up);
    return *this;
  }
  else                                          // 0 in this
  {
    if (i.inf() >= 0) {                         // i>=0
      T up = sup() * i.sup();
      r.set(FE_DOWNWARD);
      assign(inf() * i.sup(), up);
      return *this;
    }
    if (i.sup() <= 0) {                         // i<=0
      T up = inf() * i.inf();
      r.set(FE_DOWNWARD);
      assign(sup() * i.inf(), up);
      return *this;
    }
                                                  // 0 in i
    T up = std::max(inf() * i.inf(), sup() * i.sup());
    r.set(FE_DOWNWARD);
    assign( std::max(inf() * i.sup(), sup() * i.inf()), up);
    return *this;
  }
}

template < typename T >
interval<T>&
interval<T>::operator/=(interval<T> const& i)
{
  if (empty() || i.empty()) {
    *this = empty();
    return *this;
  }
  detail::rounding<T> r(FE_UPWARD);
  if (i.inf() > 0)                      // i>0
  {
    // this>=0     [inf()/sup(b); sup()/inf(b)]
    // this<=0     [inf()/inf(b); sup()/sup(b)]
    // this~=0     [inf()/inf(b); sup()/inf(b)]
    T aa = i.sup(), bb = i.inf();
    if (inf() < 0)
    {
        aa = bb;
        if (sup() < 0)
            bb = i.sup();
    };
    T up = sup() / bb;
    r.set(FE_DOWNWARD);
    assign(inf() / aa, up);
    return *this;
  }
  else if (i.sup() < 0)                 // i<0
  {
    // this>=0     [sup()/sup(b); inf()/inf(b)]
    // this<=0     [sup()/inf(b); inf()/sup(b)]
    // this~=0     [sup()/sup(b); inf()/sup(b)]
    T aa = i.sup(), bb = i.inf();
    if (inf() < 0)
    {
        bb = aa;
        if (sup() < 0)
            aa = i.inf();
    };
    T up = inf() / bb;
    r.set(FE_DOWNWARD);
    assign(sup() / aa, up);
    return *this;
  }
  else {                               // 0 in i
    if (i.inf() == 0) {
      if (i.sup() == 0) {
        *this = empty();
        return *this;
      } else if (inf() >= 0) {
        r.set(FE_DOWNWARD);
        assign(sup() / i.sup(), std::numeric_limits< T >::infinity());
        return *this;
      } else if (sup() <= 0) {
        assign(-std::numeric_limits< T >::infinity(), inf() / i.sup());
        return *this;
      }
    } else if (i.sup() == 0) {
      if (inf() >= 0) {
        assign(-std::numeric_limits< T >::infinity(), inf() / i.inf());
        return *this;
      } else if (sup() <= 0) {
        r.set(FE_DOWNWARD);
        assign(sup() / i.inf(), std::numeric_limits< T >::infinity());
        return *this;
      }
    }
    *this = whole();
    return *this;
  }
}


template < typename T > inline
interval<T>&
interval<T>::operator+=(T const& t)
{
  *this += interval<T>(t);
  return *this;
}

template < typename T > inline
interval<T>&
interval<T>::operator-=(T const& t)
{
  *this -= interval<T>(t);
  return *this;
}

template < typename T > inline
interval<T>&
interval<T>::operator*=(T const& t)
{
  *this *= interval<T>(t);
  return *this;
}

template < typename T > inline
interval<T>&
interval<T>::operator/=(T const& t)
{
  *this /= interval<T>(t);
  return *this;
}

template < typename T >
interval<T>
sqrt(interval<T> const& i)
{
  if (i.empty() || sup(i) < 0)
    return interval<T>();

  using std::sqrt;

  detail::rounding<T> r(FE_UPWARD);
  T up = sqrt(sup(i));
  r.set(FE_DOWNWARD);
  return interval<T>(sqrt(std::max(inf(i), 0)), up);
}

template < typename T >
interval<T>
square(interval<T> const& i)
{
  if (i.empty())
    return i;
  detail::rounding<T> r(FE_UPWARD);
  if (inf(i) >= 0) {
    T up = sup(i) * sup(i);
    r.set(FE_DOWNWARD);
    return interval<T>(inf(i) * inf(i), up);
  }
  if (sup(i) <= 0) {
    T up = inf(i) * inf(i);
    r.set(FE_DOWNWARD);
    return interval<T>(sup(i) * sup(i), up);
  }
  T mmax = std::max(-inf(i), sup(i));
  return interval<T>(0, mmax*mmax);
}

template < typename T > inline
interval<T>
abs(const interval<T> & i)
{
  if (i.empty() || inf(i) >= 0) return i;
  if (sup(i) <= 0) return -i;
  return interval<T>(0, std::max(-inf(i), sup(i)));
}

template < typename T > inline
T midpoint(interval<T> const& i)
{
  assert(!i.empty()); // TODO: throw?
  return (inf(i) + sup(i))/2;
}

template < typename T > inline
T width(interval<T> const& i)
{
  detail::rounding<T> r(FE_UPWARD);
  return sup(i) - inf(i);
}


// I/O
// ---

template < typename T, typename charT, typename traits >
::std::basic_ostream<charT, traits> &
operator<< (::std::basic_ostream<charT, traits> & os, interval<T> const& i)
{
  ::std::basic_ostringstream<charT, traits> s;
  s.flags(os.flags());
  s.imbue(os.getloc());
  s.precision(os.precision());
  s << "[" << i.inf() << "," << i.sup() << "]";
  return os << s.str();
}

template < typename T, typename charT, typename traits >
::std::basic_istream<charT, traits> &
operator>> (::std::basic_istream<charT, traits> & is, interval<T> & x)
{
  T inf, sup;
  charT ch;
  is >> ch;
  if (ch == '[')
  {
    is >> inf >> ch;
    if (ch == ',')
    {
      is >> sup >> ch;
      if (ch == ']')
        x = interval<T>(inf, sup);
      else
        is.setstate(::std::ios_base::failbit);
    }
    else if (ch == ']')
      x = interval<T>(inf, inf);
    else
      is.setstate(::std::ios_base::failbit);
  }
  else
  {
    is.putback(ch);
    is >> inf;
    x = interval<T>(inf, inf);
  }
  return is;
}

} // namespace std2

#endif // STD_INTERVAL_H_INCLUDED
