// Interval arithmetic proposal for C++0x.
//
// Copyright 2005  INRIA, France.
// All rights reserved.
//
// Author : Sylvain Pion.

#ifndef STD_INTERVAL_BOOL_H_INCLUDED
#define STD_INTERVAL_BOOL_H_INCLUDED

#include "interval.h"

namespace std2 {

template < typename T >
class interval;

template <>
class interval<bool>
{
  bool _i, _s;

public:

  typedef bool  value_type;

  interval();
  interval(bool t);
  interval(bool i, bool s);

private:
  void assign(bool a, bool b);
  void assign_inf(bool a);
  void assign_sup(bool a);
  
public:
  bool empty() const;
  bool inf() const;
  bool sup() const;

  operator bool() const;

  static interval indeterminate();
};


// Constructors
// ------------

inline
interval<bool>::interval()
  : _i(true), _s(false) {}

inline
interval<bool>::interval(bool b)
  : _i(b), _s(b) {}

inline
interval<bool>::interval(bool i, bool s)
  : _i(i), _s(s) {}


// Access functions
// ----------------

inline
bool interval<bool>::inf() const
{
  return _i;
}

inline
bool interval<bool>::sup() const
{
  return _s;
}

inline
bool inf(interval<bool> const& i)
{
  return i.inf();
}

inline
bool sup(interval<bool> const& i)
{
  return i.sup();
}


// Basic functions
// ---------------

inline
bool interval<bool>::empty() const
{
  return inf() == true && sup() == false;
}

inline
interval<bool>
interval<bool>::indeterminate()
{
  return interval<bool>(false, true);
}

inline
bool is_indeterminate(interval<bool> const& a)
{
  return a.inf() == false && a.sup() == true;
}


// Conversion to bool
// ------------------

inline
interval<bool>::operator bool() const
{
  if (is_singleton(*this))
    return this->inf();
  throw std::range_error("undecidable conversion of interval<bool>");
}


// Boolean operations
// ------------------

inline
interval<bool> operator!(interval<bool> const& a)
{
  return interval<bool>(!a.sup(), !a.inf());
}

inline
interval<bool> operator||(interval<bool> const& a, interval<bool> const& b)
{
  if (a.empty() || b.empty())
    return interval<bool>();
  return interval<bool>(a.inf() || b.inf(), a.sup() || b.sup());
}

inline
interval<bool> operator||(bool a, interval<bool> const& b)
{
  if (b.empty())
    return interval<bool>();
  return interval<bool>(a || b.inf(), a || b.sup());
}

inline
interval<bool> operator||(interval<bool> const& a, bool b)
{
  if (a.empty())
    return interval<bool>();
  return interval<bool>(a.inf() || b, a.sup() || b);
}

inline
interval<bool> operator&&(interval<bool> const& a, interval<bool> const& b)
{
  if (a.empty() || b.empty())
    return interval<bool>();
  return interval<bool>(a.inf() && b.inf(), a.sup() && b.sup());
}

inline
interval<bool> operator&&(bool a, interval<bool> const& b)
{
  if (b.empty())
    return interval<bool>();
  return interval<bool>(a && b.inf(), a && b.sup());
}

inline
interval<bool> operator&&(interval<bool> const& a, bool b)
{
  if (a.empty())
    return interval<bool>();
  return interval<bool>(a.inf() && b, a.sup() && b);
}


// Equality operators

inline
interval<bool> operator==(interval<bool> const& a, interval<bool> const& b)
{
  if (a.empty() || b.empty())
    return interval<bool>();
  if (std2::is_indeterminate(a) || std2::is_indeterminate(b))
    return interval<bool>::indeterminate();
  return a.inf() == b.inf();
}

inline
interval<bool> operator==(interval<bool> const& a, bool b)
{
  if (a.empty())
    return interval<bool>();
  if (std2::is_indeterminate(a))
    return interval<bool>::indeterminate();
  return a.inf() == b;
}

inline
interval<bool> operator==(bool a, interval<bool> const& b)
{
  if (b.empty())
    return interval<bool>();
  if (std2::is_indeterminate(b))
    return interval<bool>::indeterminate();
  return a == b.inf();
}

inline
interval<bool> operator!=(interval<bool> const& a, interval<bool> const& b)
{
  return ! (a == b);
}

inline
interval<bool> operator!=(interval<bool> const& a, bool b)
{
  return ! (a == b);
}

inline
interval<bool> operator!=(bool a, interval<bool> const& b)
{
  return ! (a == b);
}


} // namespace std2

#endif // STD_INTERVAL_BOOL_H_INCLUDED
