
#include "interval_cmath.h"
#include "crlibm.h"
#include "mpfi.h"

#include <cassert>
#include <limits>

namespace helper {

struct crlibm_context {
  crlibm_context(): flag(crlibm_init()) {}
  ~crlibm_context() { crlibm_exit(flag); }
 private:
  unsigned long long flag;
};

template <class T>
::std2::interval<T> mpfi_helper_tmpl(const ::std2::interval<T>& x,
  int (*f)(mpfi_t, const mpfi_t), T (*g)(const mpfr_t, mpfr_rnd_t) )
{
  // assume floating point binary representation with implicit bit
  assert( ::std::numeric_limits<T>::radix == 2 );
  const int digits = ::std::numeric_limits<T>::digits +1;
  if (x.empty())
    return x;
  mpfi_t y;
  mpfi_init2(y, digits);
  mpfi_interv_d(y, x.inf(), x.sup());
  (*f)(y, y);
  if (mpfi_is_empty(y)) {
    mpfi_clear(y);
    return ::std2::interval<T>();
  }
  mpfr_t yi, ys;
  mpfr_init2(yi, digits); mpfi_get_left (yi, y);
  mpfr_init2(ys, digits); mpfi_get_right(ys, y);
  mpfi_clear(y);
  ::std2::interval<T> result(g(yi, GMP_RNDD), g(ys, GMP_RNDU));
  mpfr_clear(yi);
  mpfr_clear(ys);
  return result;
}

inline ::std2::interval<double> mpfi_helper(const ::std2::interval<double>& x, int (*f)(mpfi_t, const mpfi_t))
{
  return mpfi_helper_tmpl(x, f, mpfr_get_d);
}


static ::std2::interval<long double> mpfi_helper(const ::std2::interval<long double>& x, int (*f)(mpfi_t, const mpfi_t))
{
  return mpfi_helper_tmpl(x, f, mpfr_get_ld);
}

} // namespace helper

namespace std2 {

  // no float specialization due to lack of support by mpfr
  
  // double specialization

template<> interval<double> acos(const interval<double>& x)
{
  interval<double> unit(-1.,1.);
  return ::helper::mpfi_helper(intersect(x,unit), mpfi_acos);
}

template<> interval<double> asin(const interval<double>& x)
{
  interval<double> unit(-1.,1.);
  return ::helper::mpfi_helper(intersect(x,unit), mpfi_asin);
}

template<> interval<double> atan(const interval<double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<double>(atan_rd(x.inf()), atan_ru(x.sup()));
}

template<> interval<double> cos(const interval<double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_cos);
}

template<> interval<double> cosh(const interval<double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() >= 0)
    return interval<double>(cosh_rd(x.inf()), cosh_ru(x.sup()));
  else if (x.sup() <= 0)
    return interval<double>(cosh_rd(x.sup()), cosh_ru(x.inf()));
  else
    return interval<double>(1, cosh_ru(std::max(x.inf(), x.sup())));
}

template<> interval<double> exp(const interval<double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<double>(exp_rd(x.inf()), exp_ru(x.sup()));
}

template<> interval<double> log(const interval<double>& x)
{
  if (x.empty() || x.sup() <= 0)
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() <= 0)
    return interval<double>(-std::numeric_limits<double>::infinity(), log_ru(x.sup()));
  else
    return interval<double>(log_rd(x.inf()), log_ru(x.sup()));
}

template<> interval<double> log10(const interval<double>& x)
{
  if (x.empty() || x.sup() <= 0)
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() <= 0)
    return interval<double>(-std::numeric_limits<double>::infinity(), log10_ru(x.sup()));
  else
    return interval<double>(log10_rd(x.inf()), log10_ru(x.sup()));
}

template<> interval<double> pow(const interval<double>&, int);
template<> interval<double> pow(const interval<double>&, const double&);
template<> interval<double> pow(const interval<double>&, const interval<double>&);
template<> interval<double> pow(const double&, const interval<double>&);

template<> interval<double> sin(const interval<double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_sin);
}

template<> interval<double> sinh(const interval<double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<double>(sinh_rd(x.inf()), sinh_ru(x.sup()));
}

template<> interval<double> tan(const interval<double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_tan);
}

template<> interval<double> tanh(const interval<double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_tanh);
}

  // long double specialization

template<> interval<long double> acos(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_acos);
}

template<> interval<long double> asin(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_asin);
}

template<> interval<long double> atan(const interval<long double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<long double>(atan_rd(x.inf()), atan_ru(x.sup()));
}

template<> interval<long double> cos(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_cos);
}

template<> interval<long double> cosh(const interval<long double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() >= 0)
    return interval<long double>(cosh_rd(x.inf()), cosh_ru(x.sup()));
  else if (x.sup() <= 0)
    return interval<long double>(cosh_rd(x.sup()), cosh_ru(x.inf()));
  else
    return interval<long double>(1, cosh_ru(std::max(x.inf(), x.sup())));
}

template<> interval<long double> exp(const interval<long double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<long double>(exp_rd(x.inf()), exp_ru(x.sup()));
}

template<> interval<long double> log(const interval<long double>& x)
{
  if (x.empty() || x.sup() <= 0)
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() <= 0)
    return interval<long double>(-std::numeric_limits<long double>::infinity(), log_ru(x.sup()));
  else
    return interval<long double>(log_rd(x.inf()), log_ru(x.sup()));
}

template<> interval<long double> log10(const interval<long double>& x)
{
  if (x.empty() || x.sup() <= 0)
    return x;
  ::helper::crlibm_context cc;
  if (x.inf() <= 0)
    return interval<long double>(-std::numeric_limits<long double>::infinity(), log10_ru(x.sup()));
  else
    return interval<long double>(log10_rd(x.inf()), log10_ru(x.sup()));
}

template<> interval<long double> pow(const interval<long double>&, int);
template<> interval<long double> pow(const interval<long double>&, const long double&);
template<> interval<long double> pow(const interval<long double>&, const interval<long double>&);
template<> interval<long double> pow(const long double&, const interval<long double>&);

template<> interval<long double> sin(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_sin);
}

template<> interval<long double> sinh(const interval<long double>& x)
{
  if (x.empty())
    return x;
  ::helper::crlibm_context cc;
  return interval<long double>(sinh_rd(x.inf()), sinh_ru(x.sup()));
}

template<> interval<long double> tan(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_tan);
}

template<> interval<long double> tanh(const interval<long double>& x)
{
  return ::helper::mpfi_helper(x, mpfi_tanh);
}

} // namespace std2
