// Example program for interval<> illustrating usage for solving functions,
// i.e. locating the zeros in a given interval. In this example, the function
// may be passed by pointer or by function object, but must not contain
// control structures (if, for, while) whose conditions depend on interval
// comparisons (due to the presence of exceptions).
//
// Sylvain Pion (solver), Herve Bronnimann (expression parsing) 2005.

// TODO :
// - What to recommend if the function uses "<" ?
//   Maybe one way would be to capture the idiom :
//     if (x >0)  return f();
//     else       return g();
//   And turn it into :
//     return hull( f(positive_part(x)), g(negative_part(x));
//   ?  Or something like that ?
// - Other way to do it is to require an interval overload
//   which guarantees the inclusion property (and leave the 
//   responsability to the user).
// - Note: it's possible to check the validity at compile-time
//   of f if it is templated by T, as in the example my_f below
//   by passing an expression type (which can also be used to
//   derive the conversion to string for output operation)

#include <vector>
#include <sstream>

#include "interval.h"
#include "interval_cmath.h"
#include "interval_solver.h"

using std::cin;
using std::cout;
using std::string;
using std2::interval;
using interval_solver::solve;
using interval_solver::merge_intervals;

// The function to execute is done via parsing an expression tree
// alternately, one may define a functor to pass to solve and bypass spirit

#include "interval_spirit_expr.h"

// Interval solver: read an expression, and output its zeroes in an interval

int main()
{
  cout.precision(20);

  typedef interval<double>  I;
  typedef std::vector<I>    Solutions;
  
  double inf = -1.0, sup = 1.0;
  double epsilon = 0.0000001;

  cout << "//////////////////////////////////////////////////////////\n\n";
  cout << "\t\tAn interval solver...\n\n";
  cout << "//////////////////////////////////////////////////////////\n\n";

  // variable name
  string var;
  cout << "enter the variable name (no spaces, beginning with alpha,\n   followed by any number of alpha/num/_): ";
  getline(cin, var);
  
  // look in tree_calc_grammar for the definition of calculator
  cout << "\ntype an expression of this variable ...or [q or Q] to quit\n";
  cout << "for instance, enter: (t*t-2)*(t-3)*(t-3)*(t-6)*t*t*(t+6)*(t+6)\n";
  cout << "expr: ";
  bool error;
  
  string str;
  while (getline(cin, str))
    {
    if (str.empty() || str[0] == 'q' || str[0] == 'Q')
      break;
    
    error = false;
    exp_tree* t = parse_exp_tree( str.c_str(), var, error );
    if (!error)
      {
      // parse and print the result
      cout << "parsing succeeded\n";
#ifdef BOOST_SPIRIT_DEBUG
      cout << "tree expression parsed as follows:\n" << *t << "\n";
      cout << "Evaluation at -2, -1, 0, 1, and 2 (for checking): "
        << eval_tree(t, -2.0) << " "
        << eval_tree(t, -1.0) << " "
        << eval_tree(t, 0.0) << " "
        << eval_tree(t, 1.0) << " "
        << eval_tree(t, 2.0) << "\n";
#endif

      // enter other parameters
      cout << "\nenter the bounds of the interval over which to search for zeroes:\n  ";
      do {
        cin >> inf >> sup;
        if (inf >= sup)
          cout << "please enter the lower, then the upper bound (upper > lower)\n  ";
      }  while (inf >= sup);
      cout << "\nenter the precision with which to isolate the zeroes:\n  ";
      cin >> epsilon;
      
      // use the solver
      Solutions solutions;
      interval_solver::solve_count = 0;
      solve( eval_tree_f<I>(t), std::back_inserter(solutions), I(inf, sup), epsilon );
      cout << "\nSolved with " << interval_solver::solve_count << " recursive calls ("
	   << solutions.size() << " intervals before merging)\n";
      cout << "\nSolutions (if any) lie in :\n";
      
      Solutions solutions2;
      merge_intervals(solutions.begin(), solutions.end(),
                      std::back_inserter(solutions2));
      std::swap(solutions, solutions2);
      
      for (Solutions::const_iterator it = solutions.begin(), end = solutions.end();
           it != end; ++it)
        {
        std::cout << *it << std::endl;
        }
      
      }
    else
      {
      cout << "parsing failed\nTry again... enter an expression, or ['q' or 'Q'] to quit\nexpr: ";
      }
    }
  
  return 0;
}
