// Example program for interval<> illustrating solving multivariate functions.
//
// Sylvain Pion, 2005.

#include <vector>
#include <algorithm>
#include <boost/array.hpp>

#include "interval.h"
#include "interval_cmath.h"
#include "interval_solver.h"

using std::cin;
using std::cout;
using std::vector;
using std::string;
using std2::interval;
using interval_solver::solve;

// The functions to execute are done via parsing an expression tree
// alternately, one may define a functor to pass to solve and bypass spirit

#include "interval_spirit_expr.h"

// Interval solver: read a system of expressions, and output its zeroes in an interval product

int main()
{
  std::cout.precision(20);
  
  typedef interval<double>  I;
#ifndef FIXED_DIM
  typedef vector<I>         A;
#else
  typedef array<I, FIXED_DIM> A;
#endif
  typedef vector<A>         Solutions;
  
  double inf = -1.0, sup = 1.0;
  double eps = 0.000001;
  int i;

  cout << "//////////////////////////////////////////////////////////\n\n";
  cout << "\t\tA multi-dimensional interval solver...\n\n";
  cout << "//////////////////////////////////////////////////////////\n\n";
      
  // variable names
  int n = 0;
  string nvar, var;
#ifndef FIXED_DIM
  cout << "enter the number of variables: ";
  do {
    getline(cin, nvar); n = atoi(nvar.c_str());
    if (n <= 0)
      cout << "Number " << n << " not a positive number, please reenter number of variables: ";
  } while (n <= 0);
#else
  int n = FIXED_DIM;
#endif
  
  vector<string> vars(n);
  cout << "enter the variable names (no spaces, one on each line, each name\n"
       << "   beginning with alpha followed by any number of alpha/num/_):\n";

  for (i=0; i<n; ++i)
#if 0
  {
    cout << "variable " << i << ": "; getline(cin, var);
    vars[i] = var;
  }
#endif
    do {
      cout << "variable " << i << ": "; getline(cin, var);
      if (i>0 && std::find( vars.begin(), vars.begin()+i, var) != vars.begin()+i)
        cout << "name already used, please enter another name\n";
      else
        vars[i] = var;
    } while (i>0 && std::find( vars.begin(), vars.begin()+i, var) != vars.begin()+i);
  
  // look in tree_calc_grammar for the definition of calculator
  int m = 0;
  string neqs;
  cout << "enter the number of equations of the system... or [q or Q] to quit\n";
  do {
    getline(cin, neqs);
    if (neqs.empty() || neqs[0] == 'q' || neqs[0] == 'Q')
      break;
    m = atoi(neqs.c_str());
    if (m <= 0)
      cout << "not a positive number, please reenter number of equations: ";
  } while (m <= 0);
  
  if (m != 0) {
    
    vector<exp_tree*> eqs(m);
#ifdef FIXED_DIM
    A box;
#else
    A box(m, I(inf, sup));
#endif
    
    cout << "\ntype " << m << " expressions of the variables ...";
    cout << "for instance, enter: x*x + y*y - 4,  and (x-1)*(x-1) + (y-1)*(y-1) -4\n";
    
    bool error;
    string str;
    for (i=0; i<m; ++i)
      do
        {    
          cout << "expr: ";
          getline(cin, str); error = false;
          exp_tree* t= parse_exp_tree( str.c_str(), vars, error );
          if (error)
            cout << "parsing failed, reenter an expression of the variables ...\n";
          else {
            cout << "parsing succeeded\n";
#ifdef BOOST_SPIRIT_DEBUG
            cout << "tree expression parsed as follows:\n" << *t << "\n";
#endif
            eqs[i] = t;
          }
        } while (error);
    
    // enter other parameters
    cout << "\nenter the bounds of the interval box over which to search for zeroes:\n";
    double epsilon;
    
    for (i=0; i<m; ++i)
      do 
        {
          cout << "dim " << i << ": ";
          cin >> inf >> sup;
          if (inf >= sup)
            cout << "please enter the lower, then the upper bound (upper > lower)\n  ";
          else
            box[i] = I(inf, sup);
        } while (inf >= sup);
    
    cout << "\nenter the precision with which to isolate the zeroes:\n  ";
    cin >> epsilon;
      
    // use the solver
    Solutions solutions;
    interval_solver::solve_count = 0;
    solve( eval_tree_d_f<I>(eqs), std::back_inserter(solutions), box, epsilon );
    cout << "\nSolved with " << interval_solver::solve_count << " recursive calls\n";
    cout << "\nSolutions (if any) lie in :\n";
    
    for (Solutions::const_iterator it = solutions.begin(), end = solutions.end();
         it != end; ++it)
      {
      for(size_t s=0; s<it->size(); ++s)
        std::cout << (*it)[s] << " ";
      std::cout << std::endl;
      }
    
  }
  
  return 0;
}
