// Interval analysis solver for systems of multivariate functions.
// It uses only basic dichotomy techniques.
//
// Copyright 2005  INRIA, France.
//
// Author : Sylvain Pion, Herve Bronnimann

// TODO :
// - See interval_solve.C's TODO
// - How to best merge blocks in order to limit theit number ?
//   The 1-dimension solution does not work, so probably the easiest
//   is for solve() to return a bool that says whether its block
//   is a full solution. (?)
//   XXX : actually, it's not clear that the 1-dim solution cannot be adapted...

#ifndef INTERVAL_SOLVER_H_INCLUDED
#define INTERVAL_SOLVER_H_INCLUDED

#include <vector>
#include <boost/array.hpp>
#include "interval.h"

namespace interval_solver {

using std::vector;
using std::size_t;
using boost::array;
using std2::interval;

namespace detail {

// Some functions to abstract away vector<interval<> > versus interval<>.

// contains_zero() is an accessory function that tests whether
// 0 (d-dim) is contained in the interval.
template < typename T >
bool contains_zero(interval<T> const& t)
{
  return contains(t, T(0));
}

template < typename T,size_t n >
bool contains_zero(array<interval<T>, n> const& t)
{
  for (size_t s=0; s<n; ++s)
    if (! contains_zero(t[s]))
      return false;
  return true;
}

template < typename T >
bool contains_zero(vector<interval<T> > const& t)
{
  for (size_t s=0; s<t.size(); ++s)
    if (! contains_zero(t[s]))
      return false;
  return true;
}

template < typename T, size_t n >
T width(array<interval<T>, n> const& t)
{
  T w = 0;
  for (size_t s=0; s<n; ++s)
    w = std::max(w, width(t[s]));
  return w;
}

template < typename T >
T width(vector<interval<T> > const& t)
{
  T w = 0;
  for (size_t s=0; s<t.size(); ++s)
    w = std::max(w, width(t[s]));
  return w;
}

template < typename T, size_t n > inline
bool contains(array<interval<T>, n> const& i1, array<interval<T>, n> const& i2)
{
  for (size_t s=0; s<n; ++s)
    if (! contains(i1[s], i2[s]))
      return false;
  return true;
}

template < typename T > inline
bool contains(vector<interval<T> > const& i1, vector<interval<T> > const& i2)
{
  assert( i1.size() == i2.size() );
  for (size_t s=0; s<i1.size(); ++s)
    if (! contains(i1[s], i2[s]))
      return false;
  return true;
}

template < typename T > inline
void assign_box(interval<T> & i, T const& t)
{
  assert( t >=0 );
  i = interval<T>(-t,t);
}

template < typename T, size_t n > inline
void assign_box(array<interval<T>, n> & i, T const& t)
{
  assert( t >= 0 );
  i.assign(i.size(), interval<T>(-t,t));
}

template < typename T > inline
void assign_box(vector<interval<T> > & i, T const& t)
{
  assert( t >= 0 );
  i.assign(i.size(), interval<T>(-t,t));
}

} // namespace detail

// Univariate solver :
// -------------------

int solve_count; // count the number of recursive calls

// Returns a sorted set of intervals (sub-intervals of current),
// which might contain zeros of f.
// The dichotomy is stopped when the width of subintervals is <= precision.
template < typename Function, typename OutputIterator, typename T >
OutputIterator
solve(Function f, OutputIterator oit,
      interval<T> const& current,
      T const& precision = 0)
{
  using namespace detail;
  ++solve_count;

  typedef typename Function::result_type R;
  typedef interval<T> I;

  R res = f(current); // Evaluate f() over current interval.

  // If 0 is not in the resulting interval, there is no solution here.
  if (! contains_zero(res))
    return oit;

  // Stop the dichotomy if res is small enough
  // (this prevents (most probably) useless work).
  // Also stop if we have reached the maximal precision.
  R r;
  assign_box(r, std::numeric_limits<T>::min());
  if (contains(r, res) || width(current) <= precision) {
    *oit++ = current;
    return oit;
  }

  // Do the dichotomy recursively.
  std::pair<I, I> ip = bisect(current);

  // Stop if we can't dichotomize anymore.
  if (is_singleton(ip.first) || is_singleton(ip.second)) {
    *oit++ = current;
    return oit;
  }

  oit  = solve(f, oit, ip.first,  precision);
  return solve(f, oit, ip.second, precision);
}


// Multivariate solver :
// ---------------------

// Returns a set of block-intervals (sub-blocks of current),
// which might contain zeros of f.
// The dichotomy is stopped when the width of subintervals is <= precision.
template < typename Function, typename OutputIterator, typename T,
           typename IntervalArray >
OutputIterator
solve(Function f, OutputIterator oit,
      IntervalArray const& current,
      T const& precision = 0)
{
  using namespace detail;
  ++solve_count;

  typedef typename Function::result_type R;
  typedef interval<T> I;
  typedef IntervalArray A;
  
  R res = f(current); // Evaluate f() over current interval.

  // If 0 is not in the resulting interval, there is no solution here.
  if (! contains_zero(res))
    return oit;

  // Stop the dichotomy if res is small enough
  // (this prevents (most probably) useless work).
  // Also stop if we have reached the maximal precision.
  R r(current); // initialize dimension
  assign_box(r, std::numeric_limits<T>::min());
  if (contains(r, res) || width(current) <= precision) {
    *oit++ = current;
    return oit;
  }

  // Otherwise bisect along every dimension
  A begin(current), end(current);
  for (size_t s=0; s<current.size(); ++s) {
    std::pair<I,I> p = bisect(current[s]);
    begin[s] = p.first; end[s] = p.second;
    // Stop if we hit a singleton along any dimension
    if (is_singleton(begin[s]) || is_singleton(end[s])) {
      *oit++ = current;
      return oit;
    }
  }

  // We could use a Hilbert curve (Hamilton path ?) ?
  // Or just binary enumeration...
  // Hilbert would be easier for merging since consecutive ?
  A it(begin);
  while (true)
  {
    // Solve recursively
    oit = solve(f, oit, it, precision);
    // Do the ++
    for (size_t s=0; s<current.size(); ++s)
    {
      // if (begin[s] >= mid_point[s]) { // Does not work !
      if (inf(it[s]) >= sup(begin[s])) {
        if (s == current.size()-1)
          return oit; // begin == end
        it[s] = begin[s];
      }
      else
      {
        it[s] = end[s];
        break;
      }
    }
  }
}


// Merge consecutive 1-d intervals from a sorted sequence.
template < typename InputIterator, typename OutputIterator >
OutputIterator
merge_intervals(InputIterator begin, InputIterator end, OutputIterator oit)
{
  typedef typename std::iterator_traits<InputIterator>::value_type I;
  typedef typename I::value_type T;
  while (begin != end)
  {
    T inf = (*begin).inf();
    T sup = (*begin).sup();
    while ( ++begin != end && sup == (*begin).inf() )
      sup = (*begin).sup();
    *oit++ = I(inf,sup);
  }
  return oit;
}

} // interval_solver

#endif // INTERVAL_SOLVER_H_INCLUDED
