//
// Copyright Herve Bronnimann (expression parsing) 2005.

// TODO
// - Integral exponents (expr^2, etc.) (DONE)
// - Identifiers to support variable and environment (e.g. interval constants) (DONE)
// - Identifiers to support function calls (DONE)
// - Support for <cmath> functions

#include <cmath>
#include <cassert>
#include <cstdlib>
#include <stack>
#include <vector>
#include <string>
#include <algorithm>
#include <iostream>

using std::stack;
using std::vector;
using std::string;

#define BOOST_SPIRIT_SINGLE_GRAMMAR_INSTANCE
#include <boost/array.hpp>
#include <boost/spirit.hpp>

using boost::array;
using namespace boost::spirit;

// Expression tree

enum { varID = 1, intID, valueID, funcID, negID, plusID, minusID, multID, divID, powID };

struct exp_tree
{
  exp_tree (string const&, int i)
  : op(varID), int_value(i)
  {}
  exp_tree (int x)
  : op(intID), int_value(x)
  {}
  exp_tree (double x)
  : op(valueID), dbl_value(x)
  {}
  exp_tree (int opID, exp_tree* l, exp_tree* r = NULL)
  : op(opID), left_child(l), right_child(r)
  {}
  
  ~exp_tree()
  {
    switch (the_op()) {
      case funcID:
      case plusID:
      case minusID:
      case multID:
      case divID:
      case powID:
        delete right();
      case negID:
        delete left();
      case valueID:
      case varID:
        break;
      default:
        assert( false );
    } 
  }
  
  int       the_op()   const { return op; }
  int       ivalue()  const { return int_value; }
  double    value()    const { return dbl_value; }
  exp_tree* left()     const { return left_child; }
  exp_tree* right()    const { return right_child; }

protected:
  template <class OS> 
  OS& print(OS& os, string const& tab) const;
public:
  template <class OS>
  OS& print(OS& os) const
  {
    return print(os, "  ");
  }

private:
  int op;
  int int_value;
  double dbl_value;
  exp_tree *left_child, *right_child; 
};

template <class OS>
OS& exp_tree::print (OS& os, string const& tab) const
{
  if (the_op() == varID)
    return os << tab << "-- VAR number " << ivalue() << "\n";
  if (the_op() == intID) {
    return os << tab << "-- INT " << ivalue() << "\n";
  }
  if (the_op() == valueID) {
    return os << tab << "-- DBL " << value() << "\n";
  }
  if (the_op() == funcID) {
    os << tab << "-- FUNCTION CALL ";
    switch(left()->ivalue()) {
      case 1: os << "(square)\n"; break;
      case 2: os << "(abs)\n"; break;
#ifdef INTERVAL_CMATH_H
      case 3: os << "(acos)\n"; break;
      case 4: os << "(asin)\n"; break;
      case 5: os << "(atan)\n"; break;
      case 6: os << "(cos)\n"; break;
      case 7: os << "(cosh)\n"; break;
      case 8: os << "(exp)\n"; break;
      case 9: os << "(log)\n"; break;
      case 10: os << "(log10)\n"; break;
      case 11: os << "(sin)\n"; break;
      case 12: os << "(sinh)\n"; break;
      case 13: os << "(tan)\n"; break;
      case 14: os << "(tanh)\n"; break;
#endif
      default: os << "(UNKNOWN!)\n"; break;
    }
    return right()->print(os, tab+"    ");
  }
  if (the_op() == negID) {
    os << tab << "-- NEG\n";
    return left()->print(os, tab+"  ");
  }
  if (the_op() == plusID) {
    os << tab << "-- PLUS\n";
    return right()->print( left()->print(os, tab+"   |"), tab+"    ");
  }
  if (the_op() == minusID) {
    os << tab << "-- MINUS\n";
    return right()->print( left()->print(os, tab+"   |"), tab+"    ");
  }
  if (the_op() == multID) {
    os << tab << "-- MULT\n";
    ;
    return right()->print( left()->print(os, tab+"   |"), tab+"    ");
  }
  if (the_op() == divID) {
    os << tab << "-- DIV\n";
    return right()->print(left()->print(os, tab+"   |"), tab+"    ");
  }  
  if (the_op() == powID) {
    os << tab << "-- POW\n";
    return right()->print(left()->print(os, tab+"   |"), tab+"    ");
  }  
  return os;
}

template <class OS>
OS& operator<< (OS& os, exp_tree const& t) 
{
  return t.print(os);
}

#include <cmath>
namespace std2 {
  // because square(interval<T>) exists but not square(double)
  template <class T>
  inline double square(T const& x) { return x*x; }
  // problem inexistant for abs()
}

template <class T, class I>
T eval_tree_d( exp_tree const* root, I const& x)
{
  using namespace std; // for cmath (abs(double) etc.)
  using namespace std2; // for square(double)

  switch (root->the_op())
    {
    case funcID:
      switch( root->left()->ivalue() ) {
        case 1: return square( eval_tree_d<T,I>(root->right(),x) );
	case 2: return abs( eval_tree_d<T,I>(root->right(),x) );
#ifdef INTERVAL_CMATH_H
	case 3: return acos( eval_tree_d<T,I>(root->right(),x) );
	case 4: return asin( eval_tree_d<T,I>(root->right(),x) );
	case 5: return atan( eval_tree_d<T,I>(root->right(),x) );
	case 6: return cos( eval_tree_d<T,I>(root->right(),x) );
	case 7: return cosh( eval_tree_d<T,I>(root->right(),x) );
	case 8: return exp( eval_tree_d<T,I>(root->right(),x) );
	case 9: return log( eval_tree_d<T,I>(root->right(),x) );
	case 10: return log10( eval_tree_d<T,I>(root->right(),x) );
	case 11: return sin( eval_tree_d<T,I>(root->right(),x) );
	case 12: return sinh( eval_tree_d<T,I>(root->right(),x) );
	case 13: return tan( eval_tree_d<T,I>(root->right(),x) );
	case 14: return tanh( eval_tree_d<T,I>(root->right(),x) );
#endif
	default: assert( false );
      }
      assert( false );
    case plusID:  
      return( eval_tree_d<T,I>(root->left(),x) + eval_tree_d<T,I>(root->right(),x) );
    case minusID: 
      return( eval_tree_d<T,I>(root->left(),x) - eval_tree_d<T,I>(root->right(),x) );
    case multID:  
      return( eval_tree_d<T,I>(root->left(),x) * eval_tree_d<T,I>(root->right(),x) );
    case divID:   
      return( eval_tree_d<T,I>(root->left(),x) / eval_tree_d<T,I>(root->right(),x) );
    case powID: 
      { // compute power
        T z = T(1), y = eval_tree_d<T,I>(root->left(),x);
        int p = root->right()->ivalue(); 
        if (p<0) { p = -p; y = T(1) / y; }
        while (p>0) { z *= y; --p; }
        return z;
      }
    case negID:
      return - eval_tree_d<T,I>(root->left(),x);
    case valueID:
      return root->value();
    case varID:
      return x[root->ivalue()];
    default:
      assert( false );
    }
  return T();
}

template <class T>
T eval_tree( exp_tree const* root, T const& x)
{
  array<T,1> y = { x };
  // vector<T> y; y.push_back(x);
  return eval_tree_d<T,array<T,1> >( root, y);
}

template <class T>
struct eval_tree_f
{
  typedef T result_type;
  eval_tree_f(exp_tree* t_) : t(t_) {}
  result_type operator() (T const& x) const { return eval_tree(t, x); }
  exp_tree* t;
};

template <class T>
struct eval_tree_d_f
{
  typedef vector<T> result_type;
  eval_tree_d_f(vector<exp_tree*> t_) : t(t_) {}
  template <size_t n>
    result_type operator() (array<T,n> const& x) const
    { assert( n == t.size() ); 
      array<T,n> result;
      for (int i=0; i<n; ++i)
        result[i] = eval_tree_d(t[i], x); 
      return result;
    }
  result_type operator() (vector<T> const& x) const
    { vector<T> result(x.size());
      for (int i=0; i<x.size(); ++i)
        result[i] = eval_tree_d<T,vector<T> >(t[i], x);
      return result;
    }
  vector<exp_tree*> t;
};



////////////////////////////////////////////////////////////////////////////
//
//  Our calculator grammar
//
////////////////////////////////////////////////////////////////////////////

//  Associated semantic actions

struct exp_error
{
  exp_error(string const& name_, char const* begin_)
  : name(name_), begin(begin_) {}
  
  void operator()(char const* str, char const* end) const
  {
    std::cerr << "parse error during clause " << name << " at position " << (int)(str-begin);
  }
  
  string name;
  char const* begin;
};

struct push_var
{

  push_var(stack<exp_tree*>& eval_)
  : eval(eval_)
  {}
    
  void operator()(char const* str, char const* end) const
  {
    eval.push(new exp_tree(string(str,end), 0));
  }

  void operator()(int i) const
  {
    eval.push(new exp_tree(string(), i) );
  }
  
  stack<exp_tree*>& eval;
};

struct push_int
{
  push_int(stack<exp_tree*>& eval_)
  : eval(eval_) {}
  
  void operator()(char const* str, char const* /*end*/) const
  {
    int x = std::strtol(str, 0, 10);
    eval.push(new exp_tree(x));
  }
  
  void operator()(int i) const
  {
    eval.push(new exp_tree(i) );
  }
  
  stack<exp_tree*>& eval;
};

struct push_value
{
  push_value(stack<exp_tree*>& eval_)
  : eval(eval_) {}
  
  void operator()(double x) const
  {
    eval.push(new exp_tree(x));
  }
  
  stack<exp_tree*>& eval;
};

struct do_op1
{
  do_op1(int const& the_op, stack<exp_tree*>& eval_)
  : m_op(the_op), eval(eval_) {}
  
  void operator()(char const*, char const*) const
  {
    eval.top() = new exp_tree(m_op, eval.top());
  }
  
  int m_op;
  stack<exp_tree*>& eval;
};

do_op1
make_op1(int the_op, stack<exp_tree*>& eval)
{
  return do_op1(the_op, eval);
}

struct do_op2
{
  do_op2(int const& the_op, stack<exp_tree*>& eval_)
  : m_op(the_op), eval(eval_) {}
  
  void operator()(char const*, char const*) const
  {
    exp_tree* rhs = eval.top();
    eval.pop();
    exp_tree* lhs = eval.top();
    eval.pop();

    eval.push(new exp_tree(m_op, lhs, rhs));
  }
  
  int m_op;
  stack<exp_tree*>& eval;
};

do_op2
make_op2(int the_op, stack<exp_tree*>& eval)
{
  return do_op2(the_op, eval);
}

// Grammar per se, allowing several variables and function calls (at most 2 args)

struct calculator : public grammar<calculator>
{
  typedef vector<string>::const_iterator svector_it;
  calculator(stack<exp_tree*>& eval_, string const* var_)
  : eval(eval_), var(var_), var_begin(NULL), var_end(NULL) {}
  calculator(stack<exp_tree*>& eval_, svector_it var_b, svector_it var_e)
  : eval(eval_), var(NULL), var_begin(var_b), var_end(var_e)
  { assert( var_b != var_e ); }
  
  void set_begin(char const* begin_) const {
    begin = begin_;
  }
  
  template <typename ScannerT>
  struct definition
  {
    definition(calculator const& self)
    {
      // dynamic variable symbol table
      if (self.var_begin != self.var_end) {
        int i=0;
        for (svector_it it = self.var_begin; it != self.var_end; ++it, ++i)
          {
          variables_d.add(it->c_str(), i);
          }
        variable =
          variables_d[push_var(self.eval)];
        ;
      } else
        variable =
          lexeme_d[ str_p(self.var->c_str())[push_var(self.eval)] ];
        ;
        
      functions_d.add
        ("square", 1)
        ("abs", 2)
#ifdef INTERVAL_CMATH_H
	("acos",3)
	("asin",4)
	("atan",5)
	("cos",6)
	("cosh",7)
	("exp",8)
	("log",9)
	("log10",10)
	("sin",11)
	("sinh",12)
	("tan",13)
	("tanh",14)
#endif
        ;
      
      function = (functions_d[push_int(self.eval)] 
                  >> '(' >> expression >> ')')[make_op2(funcID, self.eval)];
            
      integer =
        lexeme_d[ (+digit_p)[push_int(self.eval)] ]
//      | eps_p[ exp_error("integer", begin) ];
      ;

      numeric =
        lexeme_d[ (ureal_p)[push_value(self.eval)] ]
//        | eps_p[ exp_error("numeric", begin) ];
      ;
      
      exponent =
        ('^' >> integer)[make_op2(powID, self.eval)]
//      | eps_p[ exp_error("exponent", begin) ]
        ;

      factor =
        variable
        |   function
        |   numeric
        |   ( '(' >> expression >> ')' )
        |   ('-' >> factor)[make_op1(negID, self.eval)]
        |   ('+' >> factor)
//      | eps_p[ exp_error("factor", begin) ];
        ;
      
      monomial =
        factor >> ( exponent | eps_p );
      ;

      term =
        monomial
        >> *(   ('*' >> monomial)[make_op2(multID, self.eval)]
                |   ('/' >> monomial)[make_op2(divID, self.eval)]
                )
//      | eps_p[ exp_error("term", begin) ];
        ;
      
      expression =
        term
        >> *(  ('+' >> term)[make_op2(plusID, self.eval)]
               |   ('-' >> term)[make_op2(minusID, self.eval)]
               )
//      | eps_p[ exp_error("expression", begin) ];
      ;
      
      // Compile with -DBOOST_SPIRIT_DEBUG to see output
      BOOST_SPIRIT_DEBUG_RULE(expression);
      BOOST_SPIRIT_DEBUG_RULE(term);
      BOOST_SPIRIT_DEBUG_RULE(factor);
      BOOST_SPIRIT_DEBUG_RULE(monomial);
      BOOST_SPIRIT_DEBUG_RULE(exponent);
      BOOST_SPIRIT_DEBUG_RULE(numeric);
      BOOST_SPIRIT_DEBUG_RULE(integer);
      BOOST_SPIRIT_DEBUG_RULE(variable);
      BOOST_SPIRIT_DEBUG_RULE(function);
    }
    
    symbols<int> variables_d, functions_d;
    rule<ScannerT> expression, term, factor, monomial, exponent, numeric, integer, variable, function;
    rule<ScannerT> const& start() const { return expression; }
  };
  
  static char const* begin; // not multi-threading-safe, but oh well...
  
  stack<exp_tree*>& eval;
  const string *var;
  svector_it var_begin, var_end;
};

char const* calculator::begin = NULL;

////////////////////////////////////////////////////////////////////////////


exp_tree*
parse_calc(string const& str, calculator const& calc, bool& error)
{
  error = false;
  exp_tree* z = NULL;
  
  char const* begin = str.c_str();
  calc.set_begin( begin );

  parse_info<> info = parse( begin, calc, space_p);
  if (info.full && calc.eval.size() == 1)
    z = calc.eval.top();
  else
    error = true;
  
  return z;
}

exp_tree*
parse_exp_tree(string const& str, string const& var, bool& error)
{
  stack<exp_tree*> eval;
  calculator calc(eval, &var); //  Our parser
  return parse_calc(str, calc, error);
}

exp_tree*
parse_exp_tree(string const& str, vector<string> const& vars, bool& error)
{
  stack<exp_tree*> eval;
  calculator  calc(eval, vars.begin(), vars.end()); //  Our parser
  return parse_calc(str, calc, error);
}

